#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include "utils.h"


void display_help()
{
    printf("Usage: drxtool [.bin/folder] (or just drag-n-drop)\n"
           "Extracts and rebuilds DRC and DRH firmware/language blobs.\n"
           "Provide a firmware/lang .bin to unpack to a folder with the same name\n"
           "Provide a folder to repack as a firmware/lang .bin\n");
    exit(-1);
}

size_t get_fsize(FILE* f)
{
    fseeko(f, 0, SEEK_END);
    size_t fsize = ftello(f);
    rewind(f);
    return fsize;
}

bool dump_to_file(void* buf, u32 len, char* fname)
{
    FILE* f_out = fopen(fname, "wb");
    if(!f_out)
        return false;
    size_t write_len = fwrite(buf, 1, len, f_out);
    fclose(f_out);
    if(write_len != len)
        return false;
    return true;
}

void* read_to_buf(u32* size, char* fname)
{
    FILE* f_in = fopen(fname, "rb");
    if(!f_in)
    {
        printf("Failed to open %s !\n", fname);
        return NULL;
    }
    
    size_t fsize = get_fsize(f_in);
    
    void* data = malloc(fsize);
    if(!data)
    {
        printf("Failed to allocate %d bytes to read %s !\n", (u32)fsize, fname);
        return NULL;
    }
    fread(data, 1, fsize, f_in);
    fclose(f_in);
    
    *size = fsize;
    return data;
}

// fast crc32 (not like it rly matters...)
uint32_t get_crc(const void* data, size_t length)
{
  uint32_t crc = 0xFFFFFFFF;
  unsigned char* current = (unsigned char*) data;
  while (length--)
  {
    crc ^= *current++;
    for (unsigned int j = 0; j < 8; j++)
      crc = (crc >> 1) ^ (crc & 1)*0xEDB88320;
  }
  return ~crc;
}

bool path_is_dir(char* path)
{
    struct stat fstat;
    if(stat(path, &fstat) != 0)
    {
        //printf("Failed to access path %s with error %d!\n", path, errno);
        return false;
    }
    return S_ISDIR(fstat.st_mode);
}

bool path_is_file(char* path)
{
    struct stat fstat;
    if(stat(path, &fstat) != 0)
    {
        //printf("Failed to access path %s with error %d!\n", path, errno);
        return false;
    }
    return S_ISREG(fstat.st_mode);
}

void build_crc_tables(u8 crc_table[0x5000], u32 blob_type, void* data, u32 datalen)
{
    u32 i;
    
    memset(crc_table, 0, 0x5000);
    
    // set blob type at crc_table + 0
    putle32(crc_table, blob_type);
    
    // calculate subcrcs. one per 0x1000 block of data, last one to end of data
    u32* subcrc_table = (u32*)&crc_table[0x1000];
    u32 num_crcs = (datalen >> 12) + 1;
    for(i = 0; i < num_crcs; i++)
    {
        u32 offs = 0x1000 * i;
        u32 size = datalen - offs < 0x1000 ? datalen - offs : 0x1000;
        u32 crc = get_crc(data + offs, size);
        putle32(&subcrc_table[i], crc);
    }
    
    // fix super crcs - four over each 0x1000 of subcrc table
    // the first 'crc' is actually device type (DRC or DRH) as the first 0xFFC
    // bytes are CRC'd at super crcs + 0xFFC
    u32* super_crcs = (u32*)&crc_table[4];
    for(i = 0; i < 4; i++)
    {
        void* crcdata = &crc_table[0x1000 + (i * 0x1000)];
        u32 crc = get_crc(crcdata, 0x1000);
        putle32(&super_crcs[i], crc);
    }
    
    // fix header crc; 0xFFC from crc_table
    u32 crc = get_crc(crc_table, 0xFFC);
    putle32(&crc_table[0xFFC], crc);
}

typedef struct
{
    u8 offset[4];
    u8 len[4];
    char name[4];
    u8 version[4];
} fw_section;

typedef struct
{
    u8 version[4];  // can also be used to determine data type
    u8 unk1[4];     // always 0x400?
    u8 unk2[4];     // always 0x4?
    u8 data_size[4];// size of hashtables + data
} blob_hdr;

#define BLOB_DRC 0x01010000
#define BLOB_DRH 0x00010000
void unpack_fw(char* path)
{
    int ret = 0;
    u32 blob_size = 0;
    u8* blob = read_to_buf(&blob_size, path);
    if(!blob)
        return;
    
    // blob must be at least PPC hdr (0x10) + 0x1000 + 0x4000
    if(blob_size < 0x5011)
    {
        printf("File too small to be valid fw blob!\n");
        return;
    }
    
    u32 blob_type = getle32(&blob[0x10]);
    if(blob_type != BLOB_DRC && blob_type != BLOB_DRH)
    {
        printf("File is not a DRC or DRH blob! Type is: %08X\n", blob_type);
        return;
    }
    char* device = (blob_type == BLOB_DRC) ? "DRC" : "DRH";
    
    // prep, create, and change to extraction dir
    char dirname[PATH_MAX];
    sprintf(dirname, "%s_extracted", path);
    if(!path_is_dir(dirname))
    {
        if(makedir(dirname))
        {
            printf("Failed to create directory %s\n", dirname);
            return;
        }
    }
    if(chdir(dirname))
    {
        printf("Failed to move into directory %s\n", dirname);
        return;
    }
    
    // try to heuristically detect language files, this should always work
    if(blob_type == BLOB_DRC && blob[0] == 0 && blob[3] == 0)
    {
        printf("Unpacking DRC language version %d part %d\n", blob[1], blob[2]);
        if(!dump_to_file(blob, 0x10, "blob_header.bin"))
            return;
        if(!dump_to_file(blob+0x10, 4, "blob_type.bin"))
            return;
        if(!dump_to_file(blob+0x5010, blob_size-0x5010, "language_data.bin"))
            return;
    }
    else // firmware
    {
        printf("Unpacking %s FW version %d.%d rev %d\n", device, blob[0], blob[1], getbe16(&blob[2]));
        if(!dump_to_file(blob, 0x10, "blob_header.bin"))
            return;
        if(!dump_to_file(blob+0x10, 4, "blob_type.bin"))
            return;
        
        // save sections, including index
        u8* fwdata = &blob[0x5010];
        u32 num_entries = getle32(&fwdata[4]) / 0x10;
        for(int i = 0; i < num_entries; i++)
        {
            fw_section *section = (fw_section*)&fwdata[i * 0x10];
            u32 offset = getle32(&section->offset);
            u32 len = getle32(&section->len);
            printf("Section %.4s: offset 0x%X, size 0x%X\n", section->name, offset, len);
            
            // save to xxxx.bin
            char fname[4+4+1];
            memset(fname, 0, sizeof(fname));
            snprintf(fname, sizeof(fname), "%.4s.bin", section->name);
            dump_to_file(fwdata + offset, len, fname);
        }
    }
    free(blob);
    printf("Done!\n");
}

void pack_fw(char* path)
{
    // strip trailing \ and /
    u32 path_len = strlen(path);
    if(path[path_len-1] == '/' || path[path_len-1] == '\\')
        path[path_len-1] = '\0';
    
    // open output file
    char fname_out[PATH_MAX];
    snprintf(fname_out, PATH_MAX, "%s.bin", path);
    FILE* f_out = fopen(fname_out, "wb");
    if(!f_out)
    {
        printf("Failed to open %s\n", fname_out);
        return;
    }
    
    if(chdir(path))
    {
        printf("Failed to move into directory %s\n", path);
        return;
    }
    
    // get our blob header and check its size
    u32 read_size = 0;
    u8* blobhdr = read_to_buf(&read_size, "blob_header.bin");
    if(!blobhdr)
        return;
    if(read_size != 0x10)
    {
        printf("Invalid blob_header.bin size! Expected 0x10, got 0x%X", read_size);
    }
    
    // get our blob header and check its size
    u32* blob_type_le = read_to_buf(&read_size, "blob_type.bin");
    if(!blob_type_le)
        return;
    if(read_size != 0x4)
    {
        printf("Invalid blob_header.bin size! Expected 0x4, got 0x%X", read_size);
        return;
    }
    // hold onto little-endian and machine-endian versions of this
    u32 blob_type = getle32(blob_type_le);
    free(blob_type_le);
    
    if(blob_type != BLOB_DRC && blob_type != BLOB_DRH)
    {
        printf("Unrecognized blob type 0x%08X\n", blob_type);
        return;
    }
    
    u8 crc_table[0x5000];
    // language data - don't pack sections
    if(blob_type == BLOB_DRC && blobhdr[0] == 0 && blobhdr[3] == 0)
    {
        printf("Packing language data...\n");
        
        // acquire the actual data from file
        u32 lang_data_size = 0;
        void* lang_data = read_to_buf(&lang_data_size, "language_data.bin");
        if(!lang_data)
            return;
        
        // fixup blob header size and write it
        putbe32(&blobhdr[12], lang_data_size + 0x5000);
        fwrite(blobhdr, 1, 0x10, f_out);
		
        build_crc_tables(crc_table, blob_type, lang_data, lang_data_size);
        fwrite(crc_table, 1, 0x5000, f_out);
        fwrite(lang_data, 1, lang_data_size, f_out);
    }
    else
    {
        printf("Packing firmware data...\n");
		
		// read our index for reference
		u32 index_size = 0;
		u8* index = read_to_buf(&index_size, "INDX.bin");
		if(!index)
			return;
		u32 index_entries = index_size / 0x10;
		
		if(index_size < 0x20)
		{
			printf("Index is too small! Should be at least 0x20 bytes (INDX + one entry\n");
			return;
		}
		
		// iterate through our index and read in its files
		// update section lengths and offsets as we go
		u8 *sections[index_entries];
		u32 section_sizes[index_entries];
		u32 section_offsets[index_entries];
		sections[0] = index;
		section_sizes[0] = index_size;
		section_offsets[0] = 0;
		for(u32 i = 1; i < index_entries; i++)
		{
            char fname[4+4+1] = {0};
			u32 cur_size = 0;
			
            fw_section *section = (fw_section*)&index[i * 0x10];
			
            snprintf(fname, sizeof(fname), "%.4s.bin", section->name);
            sections[i] = read_to_buf(&section_sizes[i], fname);
			if(sections[i] == NULL)
				return;
			
			// calculate and set section info
			section_offsets[i] = section_offsets[i-1]+section_sizes[i-1];
			putle32(&section->len, section_sizes[i]);
			putle32(&section->offset, section_offsets[i]);
		}
		
		// get total file size and construct inner payload to be checksummed
		fw_section *section = (fw_section*)&index[(index_entries-1)*0x10];
		u32 total_size = getle32(section->offset) + getle32(section->len);
		
		void* payload = malloc(total_size);
		if(!payload)
		{
			printf("Failed to alloc %d bytes for payload building!\n", total_size);
			return;
		}
		// copy sections into payload
		for(u32 i = 0; i < index_entries; i++)
		{
			memcpy(payload+section_offsets[i], sections[i], section_sizes[i]);
			free(sections[i]);
		}
		
		// create crc table, setup 
		u8 crc_table[0x5000];
        build_crc_tables(crc_table, blob_type, payload, total_size);
		
		// fixup blob header size and write it
        putbe32(&blobhdr[12], total_size + 0x5000);
        fwrite(blobhdr, 1, 0x10, f_out);
		
        fwrite(crc_table, 1, 0x5000, f_out);
        fwrite(payload, 1, total_size, f_out);
		
		free(payload);
    }
    free(blobhdr);
    fclose(f_out);
    printf("Done!\n");
}

int main(int argc, char* argv[])
{
    int i;
    
    printf("drxtool 2.0\n"
            "(C) SALT 2016\n\n");
    if(argc != 2)
        display_help();
        
    char* path = argv[1];
    if(path_is_file(path))
        unpack_fw(path);
    else if(path_is_dir(path))
        pack_fw(path);
    else
	{
        printf("Path %s is not a file or directory!\n", path);
		display_help();
	}
	
    return 0;
}
